/*-
 * Copyright (c) 2006, 2007 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CTPP2Compiler.hpp
 *
 * $CTPP$
 */
#ifndef _CTPP2_COMPILER_H__
#define _CTPP2_COMPILER_H__ 1

/**
  @file CTPP2Compiler.hpp
  @brief CTPP2 temlate-to-bytecode compiler
*/

#include <CTPP2Syntax.h>
#include <VMOpcodeCollector.hpp>

#include <map>
#include <string>

namespace CTPP // C++ Template Engine
{
// FWD
class StaticText;
class StaticData;
class CTPP2Parser;
/**
  @class CTPP2Compiler CTPP2Parser.hpp <CTPP2Parser.hpp>
  @brief CTRPP2 compiler
*/
class CTPP2Compiler
{
public:
	/**
	  @brief Constructor
	  @param oIVMOpcodeCollector - code segment collector
	  @param oISyscalls - syscalls segment
	  @param oIStaticData - static data segment
	  @param oIStaticText - static text segment
	*/
	CTPP2Compiler(VMOpcodeCollector  & oIVMOpcodeCollector,
	              StaticText         & oISyscalls,
	              StaticData         & oIStaticData,
	              StaticText         & oIStaticText);

	/**
	  @brief Print function execution result
	*/
	INT_32 OutputFnResult();

	/**
	  @brief  Send variable to standard output collector
	*/
	INT_32 OutputVariable();

	/**
	  @brief Send text data to standard output collector
	  @param vBuffer - data
	  @param iBufferLength - data size
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 OutputStaticData(CCHAR_P vBuffer, const UINT_32 & iBufferLength);

	/**
	  @brief Send integer value to standard output collector
	  @param iData - value to print
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 OutputStaticData(const INT_64 & iData);

	/**
	  @brief Send floating point value to standard output collector
	  @param sData - value to print
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 OutputStaticData(const W_FLOAT & sData);

	/**
	  @brief Change visibility scope of CDT data (for ARRAY iterating)
	  @param szScopeName - ARRAY name
	  @param iScopeNameLength - Length of ARRAY name
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 ChangeScope(CCHAR_P szScopeName, const UINT_32 & iScopeNameLength);

	/**
	  @brief Reset scope to previous CDT data (for ARRAY iterating)
	  @param iIP - instruction pointer
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 ResetScope(const UINT_32 & iIP);

	/**
	  @brief Execute system call such as HREF_PARAM, FORM_PARAM, etc
	  @param szSyscallName - syscall name
	  @param iSyscallNameLength - Length of syscall name
	  @param iArgNum - number of arguments
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 ExecuteSyscall(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength, const UINT_32 & iArgNum);

	/**
	  @brief Push variable into stack
	  @param szVariableName - variable name
	  @param iVariableNameLength - Length of variable name
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 PushVariable(CCHAR_P szVariableName, const UINT_32 & iVariableNameLength);

	/**
	  @brief Push contextual variable into stack
	  @param iTokenType - type of token
	*/
	INT_32 PushContextualVariable(const eCTPP2LoopContextVar & iTokenType);

	/**
	  @brief Push integer value into stack
	  @param iVariable - value to push
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 PushInt(const INT_64 & iVariable);

	/**
	  @brief Push Float value into stack
	  @param dVariable - value to push
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 PushFloat(const W_FLOAT & dVariable);

	/**
	  @brief Push string variable into stack
	  @param szData - data
	  @param iDataLength - data length
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 PushString(CCHAR_P szData, const UINT_32 & iDataLength);

	/**
	  @brief Pop variable from stack
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 PopVariable();

	/**
	  @brief Check existence of stack variable
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 ExistStackVariable();

	/**
	  @brief Check existence of AR register variable
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 ExistARReg();

	// ////////////////////////////////////////////////////////////////////////////////

	/**
	  @brief Get system call by id
	  @param szSyscallName - syscall name
	  @param iSyscallNameLength - Length of syscall name
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 GetSyscallId(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength);

	/**
	  @brief Get last instruction number
	*/
	UINT_32 GetCodeSize() const;

	/**
	  @brief Get instruction by instruction number
	  @param iIP - instruction number
	  @return pointer to instruction or NULL if instruction does not exist
	*/
	VMInstruction * GetInstruction(const UINT_32 & iIP);

	/**
	  @brief Remove last instruction from code segment
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 RemoveInstr();

	/**
	  @brief Unconditional jump
	  @param iIP - new instruction pointer
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 UncondJump(const UINT_32 & iIP);

	/**
	  @brief Jump if NOT Equal
	  @param iIP - new instruction pointer
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 NEJump(const UINT_32 & iIP);

	/**
	  @brief Jump if Equal
	  @param iIP - new instruction pointer
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 EQJump(const UINT_32 & iIP);

	/**
	  @brief Stop execution
	  @return instruction pointer if success, -1 if any error occured
	*/
	INT_32 Halt();

	/**
	  @brief A destructor
	*/
	~CTPP2Compiler() throw();

private:
	friend class CTPP2Parser;

	/** Opcodes collector         */
	VMOpcodeCollector  & oVMOpcodeCollector;
	/** Syscalls                  */
	StaticText         & oSyscalls;
	/** Static data segment       */
	StaticData         & oStaticData;
	/** Static text segment       */
	StaticText         & oStaticText;

	/** Syscall cache             */
	std::map<std::string,  UINT_32> mSyscalls;
	/** Id of stored 0 to compare */
	UINT_32                iZeroId;
	/** Id of stored 0 to compare */
	UINT_32                iOneId;
};

} // namespace CTPP
#endif // _CTPP2_COMPILER_H__
// End.
