/*-
 * Copyright (c) 2006, 2007 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CTPP2Compiler.cpp
 *
 * $CTPP$
 */
#include <CTPP2Compiler.hpp>

#include <CTPP2Syntax.h>
#include <StaticData.hpp>
#include <StaticText.hpp>
#include <VMOpcodes.h>

namespace CTPP // C++ Template Engine
{

//
// Constructor
//
CTPP2Compiler::CTPP2Compiler(VMOpcodeCollector  & oIVMOpcodeCollector,
                             StaticText         & oISyscalls,
                             StaticData         & oIStaticData,
                             StaticText         & oIStaticText): oVMOpcodeCollector(oIVMOpcodeCollector),
                                                                 oSyscalls(oISyscalls),
                                                                 oStaticData(oIStaticData),
                                                                 oStaticText(oIStaticText)
{
	mSyscalls["__ctpp2_emitter"]      = oSyscalls.StoreData("__ctpp2_emitter",      15);
	mSyscalls["__ctpp2_output"]       = oSyscalls.StoreData("__ctpp2_output",       14);
	mSyscalls["__ctpp2_output_int"]   = oSyscalls.StoreData("__ctpp2_output_int",   18);
	mSyscalls["__ctpp2_output_float"] = oSyscalls.StoreData("__ctpp2_output_float", 20);
	mSyscalls["__ctpp2_output_text"]  = oSyscalls.StoreData("__ctpp2_output_text",  19);

	iZeroId = oStaticData.StoreInt(0);
	iOneId  = oStaticData.StoreInt(1);

	VMInstruction oVMInstruction[] =
	{
		{ SYSCALL, SYSCALL_PARAMS(mSyscalls["__ctpp2_emitter"], 0), 0 },
		{ MOV | ARG_SRC_AR | ARG_DST_HR, 0, 0}
	};

	oVMOpcodeCollector.Insert(oVMInstruction[0]);
	oVMOpcodeCollector.Insert(oVMInstruction[1]);
}

//
// Print function execution result
//
INT_32 CTPP2Compiler::OutputFnResult()
{
	return oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_AR                ,     0, 0 ));
}

//
// Send variable to standard output collector
//
INT_32 CTPP2Compiler::OutputVariable()
{
	VMInstruction oVMInstruction[] =
	{
		{ SYSCALL, SYSCALL_PARAMS(mSyscalls["__ctpp2_output"], 1), 0 }
	};

return oVMOpcodeCollector.Insert(oVMInstruction[0]);
}

//
// Send text data to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(CCHAR_P vBuffer, const UINT_32 & iBufferLength)
{
	if (iBufferLength == 0) { return 0; }

	INT_32 iId = oStaticText.StoreData(vBuffer, iBufferLength);
	iId = oStaticData.StoreInt(iId);

	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_INT,     iId, 0 ));
return  oVMOpcodeCollector.Insert(CreateInstruction(SYSCALL              ,     SYSCALL_PARAMS(mSyscalls["__ctpp2_output_text"], 1), 0 ));

}

//
// Send integer value to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(const INT_64 & iData)
{
	INT_32 iId = oStaticData.StoreInt(iData);

	VMInstruction oVMInstruction[] =
	{
		{ PUSH | ARG_SRC_INT, iId,                                                0 },
		{ SYSCALL           , SYSCALL_PARAMS(mSyscalls["__ctpp2_output_int"], 1), 0 }
	};

	oVMOpcodeCollector.Insert(oVMInstruction[0]);

return oVMOpcodeCollector.Insert(oVMInstruction[1]);
}

//
// Send floating point value to standard output collector
//
INT_32 CTPP2Compiler::OutputStaticData(const W_FLOAT & dData)
{
	INT_32 iId = oStaticData.StoreFloat(dData);

	VMInstruction oVMInstruction[] =
	{
		{ PUSH | ARG_SRC_FLOAT, iId,                                                  0 },
		{ SYSCALL             , SYSCALL_PARAMS(mSyscalls["__ctpp2_output_float"], 1), 0 }
	};

	oVMOpcodeCollector.Insert(oVMInstruction[0]);

return oVMOpcodeCollector.Insert(oVMInstruction[1]);
}

//
// Change visibility scope of CDT data (for ARRAY iterating)
//
INT_32 CTPP2Compiler::ChangeScope(CCHAR_P szScopeName, const UINT_32 & iScopeNameLength)
{
	// Remember old HR value
	// Get from DR register indirect value and store it into stack
	INT_32 iId   = oStaticText.StoreData(szScopeName, iScopeNameLength);
	INT_32 iNull = oStaticData.StoreInt(0);
//
//	PUSH     GR
//	MOVISTR  GR, HR["ScopeName"]
//	PUSH     FR
//	MOVSIZE  FR, GR.size()
//	PUSH     ER
//	MOV      ER, 0
//
//  @RET_POINT
//	CMP      ER, FR
//	JE       @END
//	MOVIREG  HR, GR[ER]
//

//
//	Loop code
//

//
//	INC      ER
//	JMP      RET_POINT
//  @END:

	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_HR                ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_GR                ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(MOVISTR | ARG_DST_GR | ARG_SRC_HR   ,   iId, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_FR                ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(MOVSIZE | ARG_DST_FR | ARG_SRC_GR   ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(PUSH    | ARG_SRC_ER                ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_ER | ARG_SRC_INT  , iNull, 0 ));

	UINT_32 iRetPoint = oVMOpcodeCollector.Insert(CreateInstruction(CMP     | ARG_SRC_ER | ARG_DST_FR   , iNull, 0 ));

	oVMOpcodeCollector.Insert(CreateInstruction(JE                                  , (UINT_32)-1, 0 ));

	oVMOpcodeCollector.Insert(CreateInstruction(MOVIREG | ARG_DST_HR | ARG_SRC_GR   , ARG_SRC_ER, 0 ));

return iRetPoint;
}

//
// Reset scope to previous CDT data (for ARRAY iterating)
//
INT_32 CTPP2Compiler::ResetScope(const UINT_32 & iIP)
{
//
//	INC      ER
//	JMP      RET_POINT
//  @END:
//      POP      ER
//      POP      FR
//      POP      GR

	oVMOpcodeCollector.Insert(CreateInstruction(INC     | ARG_SRC_ER                ,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(JMP                                 ,   iIP, 0 ));

	UINT_32 iEndPoint =
	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_ER | ARG_SRC_STACK,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, 0 ));

	VMInstruction * pInstr = oVMOpcodeCollector.GetInstruction(iIP + 1);
	if (pInstr == NULL) { throw "Ouch!"; }
	pInstr -> argument = iEndPoint;

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_FR | ARG_SRC_STACK,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, 0 ));

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_GR | ARG_SRC_STACK,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, 0 ));

	oVMOpcodeCollector.Insert(CreateInstruction(MOV     | ARG_DST_HR | ARG_SRC_STACK,     0, 0 ));
	oVMOpcodeCollector.Insert(CreateInstruction(POP                                 ,     1, 0 ));

return iEndPoint;
}

//
// Get system call by id
//
INT_32 CTPP2Compiler::GetSyscallId(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength)
{
	std::string sTMP(szSyscallName, iSyscallNameLength);
	std::map<std::string, UINT_32>::const_iterator itmSyscalls = mSyscalls.find(sTMP);
	if (itmSyscalls != mSyscalls.end()) { return itmSyscalls -> second; }

	UINT_32 iSyscallNum = oSyscalls.StoreData(szSyscallName, iSyscallNameLength);
	mSyscalls[sTMP] = iSyscallNum;

return iSyscallNum;
}

//
// Execute system call such as HREF_PARAM, FORM_PARAM, etc
//
INT_32 CTPP2Compiler::ExecuteSyscall(CCHAR_P szSyscallName, const UINT_32 & iSyscallNameLength, const UINT_32 & iArgNum)
{
return oVMOpcodeCollector.Insert(CreateInstruction(SYSCALL, SYSCALL_PARAMS(GetSyscallId(szSyscallName, iSyscallNameLength), iArgNum), 0));
}

//
// Push contextual variable into stack
//
INT_32 CTPP2Compiler::PushContextualVariable(const eCTPP2LoopContextVar & iTokenType)
{
	switch (iTokenType)
	{
//
//  N + 0:  MOV   AR, Zero
//  N + 1:  CMP   ER, AR
//  N + 2:  JODD  @N1
//  N + 3:  PUSH  Zero
//  N + 4:  JMP   @END
// @N1
//  N + 5:  PUSH  One
// @END
//  N + 6:  something...
//
		case TMPL_LOOP_FIRST:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, 0));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                       , iIP + 4, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT       , iZeroId, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                      , iIP + 5, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, 0));
			}

//
//  N + 0:  MOV   AR, ER
//  N + 1:  INC   AR
//  N + 1:  CMP   FR, AR
//  N + 2:  JE    @N1
//  N + 3:  PUSH  Zero
//  N + 4:  JMP   @END
// @N1
//  N + 5:  PUSH  One
// @END
//  N + 6:  something...
//
		case TMPL_LOOP_LAST:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             , 0, 0));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_FR | ARG_DST_AR, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                       , iIP + 4, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT       , iZeroId, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                      , iIP + 5, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, 0));
			}

//
//  N + 0:   MOV   AR, Zero  |
//  N + 1:   CMP   ER, AR    | 0
//  N + 2:   JE    @N0       | 1 -> 6
//  N + 5:   MOV   AR, ER    | 2
//  N + 6:   INC   AR        | 3
//  N + 7:   CMP   FR, AR    | 4
//  N + 8:   JN    @N1       | 5 -> 8
// N0:
//  N + 9:   PUSH  Zero      | 6
//  N + 10:  JMP   @END      | 7 -> 9
// @N1                       |
//  N + 5:   PUSH  One       | 8
// @END                      |
//  N + 6:  something...     | 9
//
		case TMPL_LOOP_INNER:
			{

				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, 0));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JE                           ,  iIP + 6, 0));

				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER,        0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             ,        0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_FR | ARG_DST_AR,        0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JN                           ,  iIP + 8, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT           ,   iZeroId, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP                          ,  iIP + 9, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT    ,   iOneId, 0));
			}

		case TMPL_LOOP_EVEN:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, 0));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JODD,               iIP + 4, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iZeroId, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP          ,      iIP + 5, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, 0));
			}

		case TMPL_LOOP_ODD:
			{
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_INT, iZeroId, 0));
				UINT_32 iIP = oVMOpcodeCollector.Insert(CreateInstruction(CMP | ARG_SRC_ER | ARG_DST_AR, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JODD,               iIP + 4, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iOneId, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(JMP          ,      iIP + 5, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iZeroId, 0));
			}

		case TMPL_LOOP_COUNTER:
			{
				// Counter
				oVMOpcodeCollector.Insert(CreateInstruction(MOV | ARG_DST_AR | ARG_SRC_ER, 0, 0));
				oVMOpcodeCollector.Insert(CreateInstruction(INC | ARG_SRC_AR             , 0, 0));
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_AR, 0, 0));
			}

		case TMPL_LOOP_SIZE:
			{
				// Loop size
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_FR     , 0, 0));
			}

		case TMPL_LOOP_CONTENT:
			{
				// Content of variable
				return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_HR     , 0, 0));
			}
		default:
			;
	}

return 0;
}

//
// Push variable into stack
//
INT_32 CTPP2Compiler::PushVariable(CCHAR_P szVariableName, const UINT_32 & iVariableNameLength)
{
	INT_32 iId = oStaticText.StoreData(szVariableName, iVariableNameLength);

//
//  N + 0:  PUSH HR["variable"]
//  N + 1:  JE N + 3
//  N + 2:  POP 1
//  N + 3:  PUSH DR["variable"]
//  N + 4:  something...
//

	// Check from HR register indirect value
	INT_32 iPos = oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_IND_STR | ARG_DST_HR, iId, 0));
	oVMOpcodeCollector.Insert(CreateInstruction(JE, iPos + 4, 0));
	oVMOpcodeCollector.Insert(CreateInstruction(POP, 1, 0));
	iPos = oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_IND_STR | ARG_DST_DR, iId, 0));

return iPos;
}

//
// Push integer value into stack
//
INT_32 CTPP2Compiler::PushInt(const INT_64 & iVariable)
{
	INT_32 iId = oStaticData.StoreInt(iVariable);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_INT, iId, 0));
}

//
// Push Float value into stack
//
INT_32 CTPP2Compiler::PushFloat(const W_FLOAT & dVariable)
{
	INT_32 iId = oStaticData.StoreFloat(dVariable);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_FLOAT, iId, 0));
}

//
// Push string value into stack
//
INT_32 CTPP2Compiler::PushString(CCHAR_P szData, const UINT_32 & iDataLength)
{
	INT_32 iId = oStaticText.StoreData(szData, iDataLength);

return oVMOpcodeCollector.Insert(CreateInstruction(PUSH | ARG_SRC_STR, iId, 0));
}

//
// Check existence of stack variable
//
INT_32 CTPP2Compiler::ExistStackVariable() { return oVMOpcodeCollector.Insert(CreateInstruction(EXIST | ARG_SRC_STACK, 0, 0)); }

//
// Check existence of AR register variable
//
INT_32 CTPP2Compiler::ExistARReg() { return oVMOpcodeCollector.Insert(CreateInstruction(EXIST | ARG_SRC_AR, 0, 0)); }

//
// Pop variable from stack
//
INT_32 CTPP2Compiler::PopVariable() { return oVMOpcodeCollector.Insert(CreateInstruction(POP, 1, 0)); }

//
// Remove last instruction from code segment
//
INT_32 CTPP2Compiler::RemoveInstr() { return oVMOpcodeCollector.Remove(); }

//
// Unconditional jump
//
INT_32 CTPP2Compiler::UncondJump(const UINT_32 & iIP) { return oVMOpcodeCollector.Insert(CreateInstruction(JMP, iIP, 0)); }

//
// Jump if NOT Equal
//
INT_32 CTPP2Compiler::NEJump(const UINT_32 & iIP) { return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_NE, iIP, 0)); }

//
// Jump if Equal
//
INT_32 CTPP2Compiler::EQJump(const UINT_32 & iIP) { return oVMOpcodeCollector.Insert(CreateInstruction(JXX | FL_EQ, iIP, 0)); }

//
//  Stop execution
//
INT_32 CTPP2Compiler::Halt() { return oVMOpcodeCollector.Insert(CreateInstruction(HLT, 0, 0)); }

//
// Get last instruction number
//
UINT_32 CTPP2Compiler::GetCodeSize() const { return oVMOpcodeCollector.GetCodeSize(); }

//
// Get instrcution by instruction number
//
VMInstruction * CTPP2Compiler::GetInstruction(const UINT_32 & iIP) { return oVMOpcodeCollector.GetInstruction(iIP); }

//
// A destructor
//
CTPP2Compiler::~CTPP2Compiler() throw()
{
	;;
}

} // namespace CTPP
// End.
