/*-
 * Copyright (c) 2006, 2007 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CTPP2Parser.cpp
 *
 * $CTPP$
 */

#include <CTPP2Parser.hpp>

#include <CTPP2SourceLoader.hpp>
#include <CTPPParserException.hpp>

#include <stdio.h>

/**
  @file VMOpcodeCollector.hpp
  @brief Collector of generated operation codes
*/

namespace CTPP // C++ Template Engine
{

//
// CTPP2 reserved words
//
CTPP2Parser::CTPP2Keyword CTPP2Parser::aKeywords [] =
{
	{ TMPL_VAR_TOK,     sizeof(TMPL_VAR_TOK),     TMPL_var     },
	{ TMPL_IF_TOK,      sizeof(TMPL_IF_TOK),      TMPL_if      },
	{ TMPL_UNLESS_TOK,  sizeof(TMPL_UNLESS_TOK),  TMPL_unless  },
	{ TMPL_ELSE_TOK,    sizeof(TMPL_ELSE_TOK),    TMPL_else    },
	{ TMPL_ELSIF_TOK,   sizeof(TMPL_ELSIF_TOK),   TMPL_elsif   },
	{ TMPL_LOOP_TOK,    sizeof(TMPL_LOOP_TOK),    TMPL_loop    },
	{ TMPL_INCLUDE_TOK, sizeof(TMPL_INCLUDE_TOK), TMPL_include },
	{ TMPL_UDF_TOK,     sizeof(TMPL_UDF_TOK),     TMPL_udf     },
	{ TMPL_COMMENT_TOK, sizeof(TMPL_COMMENT_TOK), TMPL_comment },
	{ TMPL_BREAK_TOK,   sizeof(TMPL_BREAK_TOK),   TMPL_break   },
	// Terminator
	{ NULL,             0, UNDEF }
};

//
// Loop modifiers
//
CTPP2Parser::CTPP2LoopKeyword CTPP2Parser::aLoopKeywords [] =
{
	{ TMPL_GLOBAL_VARS_TOK,      sizeof(TMPL_GLOBAL_VARS_TOK),     TMPL_USE_GLOBAL_VARS  },
	{ TMPL_CONTEXT_VARS_TOK,     sizeof(TMPL_CONTEXT_VARS_TOK),    TMPL_USE_CONTEXT_VARS },
	{ TMPL_NO_GLOBAL_VARS_TOK,   sizeof(TMPL_NO_GLOBAL_VARS_TOK),  TMPL_NO_GLOBAL_VARS   },
	{ TMPL_NO_CONTEXT_VARS_TOK,  sizeof(TMPL_NO_CONTEXT_VARS_TOK), TMPL_NO_CONTEXT_VARS  },
	// Terminator
	{ NULL,                      0, 0 }
};

//
// Loop modifiers
//
CTPP2Parser::CTPP2LoopContextVar CTPP2Parser::aLoopContextVars [] =
{
	{ TMPL_LOOP_FIRST_TOK,       sizeof(TMPL_LOOP_FIRST_TOK),     TMPL_LOOP_FIRST   },
	{ TMPL_LOOP_LAST_TOK,        sizeof(TMPL_LOOP_LAST_TOK),      TMPL_LOOP_LAST    },
	{ TMPL_LOOP_INNER_TOK,       sizeof(TMPL_LOOP_INNER_TOK),     TMPL_LOOP_INNER   },
	{ TMPL_LOOP_ODD_TOK,         sizeof(TMPL_LOOP_ODD_TOK),       TMPL_LOOP_ODD     },
	{ TMPL_LOOP_COUNTER_TOK,     sizeof(TMPL_LOOP_COUNTER_TOK),   TMPL_LOOP_COUNTER },
	{ TMPL_LOOP_EVEN_TOK,        sizeof(TMPL_LOOP_EVEN_TOK),      TMPL_LOOP_EVEN    },
	{ TMPL_LOOP_SIZE_TOK,        sizeof(TMPL_LOOP_SIZE_TOK),      TMPL_LOOP_SIZE    },
	{ TMPL_LOOP_CONTENT_TOK,     sizeof(TMPL_LOOP_CONTENT_TOK),   TMPL_LOOP_CONTENT },
	// Terminator
	{ NULL,                      0, TMPL_LOOP_UNDEF }
};

//
// Constructor
//
CTPP2Parser::CTPP2Parser(CTPP2SourceLoader  * pISourceLoader,
                         CTPP2Compiler      * pICTPP2Compiler): pSourceLoader(pISourceLoader),
                                                                pCTPP2Compiler(pICTPP2Compiler),
                                                                oSyscalls(pCTPP2Compiler -> oSyscalls),
                                                                oStaticData(pCTPP2Compiler -> oStaticData),
                                                                oStaticText(pCTPP2Compiler -> oStaticText)
{
	;;
}

//
// CTPP2 open tag ([Tt][Mm][Pp][Ll])
//
CCharIterator CTPP2Parser::IsOpenTag(CCharIterator szData, CCharIterator szEnd)
{
	CCharIterator szIter1 = szData;
	CCharIterator szIter2 = CTPP2_OPEN_TAG;

	while (szIter1 != szEnd)
	{
		if ((*szIter1 | 0x20) != *szIter2) { break; }
		++szIter1; ++szIter2;
		if (*szIter2 == '\0')
		{
			if (*szIter1 == '_') { return szIter1; }
			return NULL;
		}
	}

return NULL;
}

//
// CTPP2 close tag (/[Tt][Mm][Pp][Ll])
//
CCharIterator CTPP2Parser::IsCloseTag(CCharIterator szData, CCharIterator szEnd)
{
	CCharIterator szIter2 = CTPP2_CLOSE_TAG;

	if (szData != szEnd && *szData != '/') { return NULL; }
	++szData;

	while (szData != szEnd)
	{
		if ((*szData | 0x20) != *szIter2) { break; }
		++szData; ++szIter2;
		if (*szIter2 == '\0')
		{
			if (*szData == '_') { return szData; }
			return NULL;
		}
	}
return NULL;
}

//
// CTPP2 Operator (var | if | else | unless | loop | udf | comment)
//
CCharIterator CTPP2Parser::IsOperator(CCharIterator szData, CCharIterator szEnd, eCTPP2Operator & oOperatorType)
{
	oOperatorType = UNDEF;

	CTPP2Keyword * oKeyword = aKeywords;
	while (oKeyword -> keyword != NULL)
	{
		CCharIterator szIter1 = szData;
		CCharIterator szIter2 = oKeyword -> keyword;
		while (szIter1 != szEnd)
		{
			if ((*szIter1 | 0x20) != *szIter2) { break; }
			++szIter1; ++szIter2;
			if (*szIter2 == '\0')
			{
				oOperatorType = oKeyword -> keyword_operator;
				return szIter1;
			}
		}
		++oKeyword;
	}

return NULL;
}

//
// CTPP2 branch expression
//
CCharIterator CTPP2Parser::IsIfExpr(CCharIterator szData, CCharIterator szEnd, eCTPP2ExprOperator & eResultOperator)
{
	CCharIterator sTMP = IsExpr(szData, szEnd, eResultOperator);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected '<TMPL_if expression>'", szData.GetLine(), szData.GetLinePos()); }

	// All Okay
	szData = sTMP;

	// Skip white space
	szData = IsWhiteSpace(szData, szEnd, 0);

	// Check close symbol
	if (*szData != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", szData.GetLine(), szData.GetLinePos()); }

return szData;
}

//
// CTPP2 reverse branch expression
//
CCharIterator CTPP2Parser::IsUnlessExpr(CCharIterator szData, CCharIterator szEnd, eCTPP2ExprOperator & eResultOperator)
{
	return IsExpr(szData, szEnd, eResultOperator);
}

//
// CTPP2 loop expression
//
CCharIterator CTPP2Parser::IsLoopExpr(CCharIterator szData, CCharIterator szEnd, UINT_32 & iRetPoint)
{
	UINT_32 iLoopKeyword = 0;
	// Find modifiers
	// First loop modifier
	CCharIterator sTMP = IsLoopKeyword(szData, szEnd, iLoopKeyword);
	if (sTMP != NULL)
	{
		szData = sTMP;
		// Skip white space
		sTMP = IsWhiteSpace(szData, szEnd);
		if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
		szData = sTMP;
		// Second loop modifier
		sTMP = IsLoopKeyword(szData, szEnd, iLoopKeyword);
		if (sTMP != NULL)
		{
			szData = sTMP;

			sTMP = IsWhiteSpace(szData, szEnd);
			if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
			szData = sTMP;
		}
	}

	if ((iLoopKeyword & (TMPL_USE_GLOBAL_VARS | TMPL_NO_GLOBAL_VARS)) == (TMPL_USE_GLOBAL_VARS | TMPL_NO_GLOBAL_VARS))
	{
		throw CTPPParserSyntaxError("cannot combine '" TMPL_GLOBAL_VARS_TOK "' and '" TMPL_NO_GLOBAL_VARS_TOK "' in one loop", szData.GetLine(), szData.GetLinePos());
	}

	if ((iLoopKeyword & (TMPL_USE_CONTEXT_VARS | TMPL_NO_CONTEXT_VARS)) == (TMPL_USE_CONTEXT_VARS | TMPL_NO_CONTEXT_VARS))
	{
		throw CTPPParserSyntaxError("cannot combine '" TMPL_CONTEXT_VARS_TOK "' and '" TMPL_NO_CONTEXT_VARS_TOK "' in one loop", szData.GetLine(), szData.GetLinePos());
	}

	// Find loop name
	sTMP = IsVar(szData, szEnd);

	if (sTMP == NULL) { throw CTPPParserSyntaxError("incorrect loop condition", szData.GetLine(), szData.GetLinePos()); }

	// Remember RET_POINT
	iRetPoint = pCTPP2Compiler -> ChangeScope(szData(), sTMP() - szData());

//fprintf(stderr, "PUSH_LOOP `");
//fwrite(szData(), sTMP() - szData(), 1, stderr);
//fprintf(stderr, "`\n");

return sTMP;
}

//
// CTPP2 loop modifier (__global_vars__ | __context_vars__ | __no_global_vars__ | __no_context_vars__ )
//
CCharIterator CTPP2Parser::IsLoopKeyword(CCharIterator szData, CCharIterator szEnd, UINT_32 & iLoopKeyword)
{
	if (szData == szEnd || *szData != '_') { return NULL; }

	CTPP2LoopKeyword * oKeyword = aLoopKeywords;
	while (oKeyword -> keyword != NULL)
	{
		CCharIterator szIter1 = szData;
		CCharIterator szIter2 = oKeyword -> keyword;
		while (szIter1 != szEnd)
		{
			if (((*szIter1 | 0x20) != *szIter2) &&
			     (*szIter2 == '_' && *szIter1 != *szIter2)) { break; }
			++szIter1; ++szIter2;
			if (*szIter2 == '\0')
			{
				iLoopKeyword |= oKeyword -> keyword_operator;
				return szIter1;
			}
		}
		++oKeyword;
	}

return NULL;
}

//
// CTPP2 expression ( Function ( expression ) | string | variable | number )
//
CCharIterator CTPP2Parser::IsExpr(CCharIterator szData, CCharIterator szEnd, eCTPP2ExprOperator & eResultOperator)
{
	// Skip white space
	szData = IsWhiteSpace(szData, szEnd, 0);

	CCharIterator sFuncEnd = NULL;

	UINT_32 iFunctionParams = 0;
	// Check function
	CCharIterator sTMP = IsFunc(szData, szEnd, sFuncEnd, iFunctionParams);
	if (sTMP != NULL)
	{
		// Push syscall
		pCTPP2Compiler -> ExecuteSyscall(szData(), sFuncEnd() - szData(), iFunctionParams);
		eResultOperator = EXPR_FUNCTION;
//fprintf(stderr, "PUSH_FUNC(%d) `", iFunctionParams);
//fwrite(szData(), sFuncEnd() - szData(), 1, stderr);
//fprintf(stderr, "`\n");
		return sTMP;
	}

	eCTPP2LoopContextVar eContextVar;
	// Check context variable
	sTMP = IsContextVar(szData, szEnd, eContextVar);
	if (sTMP != NULL)
	{
		pCTPP2Compiler -> PushContextualVariable(eContextVar);
		eResultOperator = EXPR_VARIABLE;
//fprintf(stderr, "PUSH_CONTEXT_VARIABLE `");
//fwrite(szData(), sTMP() - szData(), 1, stderr);
//fprintf(stderr, "`\n");
		return sTMP;
	}

	sTMP = IsVar(szData, szEnd);
	if (sTMP != NULL)
	{
		pCTPP2Compiler -> PushVariable(szData(), sFuncEnd() - szData());
		eResultOperator = EXPR_VARIABLE;
//fprintf(stderr, "PUSH_VARIABLE `");
//fwrite(szData(), sTMP() - szData(), 1, stderr);
//fprintf(stderr, "`\n");
		return sTMP;
	}

	// Check number
	sTMP = IsNum(szData, szEnd);
	if (sTMP != NULL)
	{
		if (iParsedNumberType == 0)
		{
			pCTPP2Compiler -> PushInt(iIntData);
			eResultOperator = EXPR_INT_VALUE;
		}
		else
		{
			pCTPP2Compiler -> PushFloat(dFloatData);
			eResultOperator = EXPR_FLOAT_VALUE;
		}

//fprintf(stderr, "PUSH_NUMBER `");
//fwrite(szData(), sTMP() - szData(), 1, stderr);
//fprintf(stderr, "`\n");
		return sTMP;
	}

	// Check string
	sTMP = IsString(szData, szEnd);
	if (sTMP != NULL)
	{
		pCTPP2Compiler -> PushString(sTMPBuf.c_str(), sTMPBuf.size());
		eResultOperator = EXPR_STRING_VALUE;
//fprintf(stderr, "PUSH_STRING `%s`\n", sTMPBuf.c_str());
		return sTMP;
	}

return NULL;
}

//
// CTPP2 string "blah-blah \" clah-clah " | 'blah-blah \' clah-clah '
//
CCharIterator CTPP2Parser::IsString(CCharIterator szData, CCharIterator szEnd)
{
	sTMPBuf.erase();

	INT_32 bEscaped = 0;
	// End-of-String
	if (szData == szEnd) { return NULL; }
	CHAR_8 chEOS = *szData;
	if (chEOS != '"' && chEOS != '\'') { return NULL; }
	++szData;
	if (szData == szEnd) { throw CTPPParserSyntaxError("missing terminating character", szData.GetLine(), szData.GetLinePos()); }

	while (szData != szEnd)
	{
		// Escaper
		if (*szData == '\\' && bEscaped == 0)
		{
			++szData;
			bEscaped = 1;
		}
		// Other cases
		else
		{
			// End-of-String ?
			if (*szData == chEOS && bEscaped == 0) { return ++szData; }
			if (bEscaped == 1)
			{
				switch (*szData)
				{
					// Octal value
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					{
						UINT_32 iTMP = 0;
						for (INT_32 iI = 2; iI >= 0; --iI)
						{
							UCHAR_8 iCh = *szData;

							if (iCh >= '0' && iCh <= '7')
							{
								iTMP = (iTMP << 3);
								iTMP += (iCh - '0');
							}
							else { throw CTPPParserSyntaxError("invalid octal escape sequence", szData.GetLine(), szData.GetLinePos()); }

							++szData;
						}
						if (iTMP > 255) { throw CTPPParserSyntaxError("octal escape sequence out of range", szData.GetLine(), szData.GetLinePos()); }
						sTMPBuf += UCHAR_8(iTMP);
					}
					break;
					// Error
					case '8':
					case '9':
						throw CTPPParserSyntaxError("invalid octal escape sequence", szData.GetLine(), szData.GetLinePos());
						break;
					// Hexadecimal value
					case 'x':
					case 'X':
					{
						++szData;
						UINT_32 iTMP = 0;
						for (INT_32 iI = 1; iI >= 0; --iI)
						{
							if (szData == szEnd) { throw CTPPParserSyntaxError("invalid hexadecimal escape sequence", szData.GetLine(), szData.GetLinePos()); }

							UCHAR_8 iCh = *szData;

							if (iCh >= '0' && iCh <= '9') { iTMP += ((iCh - '0') << (iI * 4)); }
							else
							{
								iCh |= 0x20;
								if (*szData >= 'a' && *szData <= 'f') { iTMP += ((iCh - 'a' + 10) << (iI * 4)); }
								else                                  { throw CTPPParserSyntaxError("invalid hexadecimal escape sequence", szData.GetLine(), szData.GetLinePos()); }
							}

							++szData;
						}
						sTMPBuf += UCHAR_8(iTMP);
					}
					break;

					// man ascii
					case 'a': sTMPBuf += '\a'; ++szData; break;
					case 'b': sTMPBuf += '\b'; ++szData; break;
					case 't': sTMPBuf += '\t'; ++szData; break;
					case 'n': sTMPBuf += '\n'; ++szData; break;
					case 'v': sTMPBuf += '\v'; ++szData; break;
					case 'f': sTMPBuf += '\f'; ++szData; break;
					case 'r': sTMPBuf += '\r'; ++szData; break;
					// Unicode + 4 digits
					case 'u':
					{
						++szData;
						UINT_32 iTMP = 0;
						for (INT_32 iI = 3; iI >= 0; --iI)
						{
							if (szData == szEnd) { throw CTPPParserSyntaxError("invalid unicode escape sequence", szData.GetLine(), szData.GetLinePos()); }

							UCHAR_8 iCh = *szData;

							if (iCh >= '0' && iCh <= '9') { iTMP += ((iCh - '0') << (iI * 4)); }
							else
							{
								iCh |= 0x20;
								if (*szData >= 'a' && *szData <= 'f') { iTMP += ((iCh - 'a' + 10) << (iI * 4)); }
								else                                  { throw CTPPParserSyntaxError("invalid unicode escape sequence", szData.GetLine(), szData.GetLinePos()); }
							}

							++szData;
						}
						UCHAR_8 iCh = ((iTMP & 0x0000FF00) >> 8);
						if (iCh != 0) { sTMPBuf += iCh; }

						iCh = (iTMP & 0x000000FF);
						sTMPBuf += iCh;
					}
					break;
					default:
						sTMPBuf += *szData;
						++szData;
				}
			}
			else
			{
				sTMPBuf += *szData;
				++szData;
			}

			bEscaped = 0;
		}
	}

return NULL;
}

//
// CTPP2 old fashion string blah-blah.clah-clah
//
CCharIterator CTPP2Parser::IsStringCompatOldVersion(CCharIterator szData, CCharIterator szEnd)
{
	sTMPBuf.erase();
	if (szData == szEnd) { return NULL; }
	while (szData != szEnd)
	{
		if (*szData == ' ' || *szData == '\t' || *szData == TMPL_CLOSE_SYMBOL) { return szData; }
		sTMPBuf += *szData;
		szData++;
	}

return NULL;
}

//
// CTPP2 variable [a-zA-Z_][a-zA-Z0-9_]
//
CCharIterator CTPP2Parser::IsVar(CCharIterator szData, CCharIterator szEnd)
{
	if (szData != szEnd &&
	    !((*szData >= 'a' && *szData <= 'z') ||
	      (*szData >= 'A' && *szData <= 'Z') || *szData == '_')
	    )
	{
		return NULL;
	}
	++szData;

	while (szData != szEnd)
	{
		if (!((*szData >= 'a' && *szData <= 'z') ||
		      (*szData >= 'A' && *szData <= 'Z') ||
		      (*szData >= '0' && *szData <= '9') ||
		       *szData == '_')
		    )
		{
			return szData;
		}
		++szData;
	}
return NULL;
}

//
// CTPP2 Loop contextual variable (__first__ | __last__ | __inner__ | __odd__ | __even__ | __counter__ | __size__ | __content__ )
//
CCharIterator CTPP2Parser::IsContextVar(CCharIterator szData, CCharIterator szEnd, eCTPP2LoopContextVar & eContextVariable)
{
	if (szData == szEnd || *szData != '_') { return NULL; }

	CTPP2LoopContextVar * oKeyword = aLoopContextVars;
	while (oKeyword -> keyword != NULL)
	{
		CCharIterator szIter1 = szData;
		CCharIterator szIter2 = oKeyword -> keyword;
		while (szIter1 != szEnd)
		{
			if (!((*szIter1 | 0x20) == *szIter2 || (*szIter2 == '_' && *szIter1 == *szIter2))) { break; }
			++szIter1; ++szIter2;
			if (*szIter2 == '\0')
			{
				eContextVariable = oKeyword -> keyword_operator;
				return szIter1;
			}
		}
		++oKeyword;
	}

	throw CTPPParserSyntaxError("Invalid contextual variable", szData.GetLine(), szData.GetLinePos());

return NULL;
}

//
// CTPP2 number [-]0-9[. [0-9] ] [E][+- 0-9]
//
CCharIterator CTPP2Parser::IsNum(CCharIterator szData, CCharIterator szEnd)
{
	CCharIterator szStart = szData;

	// First digit
	if (szData != szEnd && !(*szData >= '0' && *szData <= '9')) { return NULL; }
	++ szData;

	// Digits
	while (szData != szEnd)
	{
		if (!(*szData >= '0' && *szData <= '9')) { break; }
		++ szData;
	}

	// Optional '.'. If not found, this is an integer value
	if (szData == szEnd || *szData != '.')
	{
		INT_32 iDigitSize = szEnd() - szStart();
		CHAR_P szDigit = new CHAR_8[iDigitSize + 1];
		memcpy(szDigit, szStart(), iDigitSize);
		szDigit[iDigitSize] = '\0';

		long long iLL = 0;
		sscanf(szDigit, "%lli", &iLL);

		iIntData = iLL;
		iParsedNumberType = 0;

		delete [] szDigit;
		return szData;
	}
	++szData;

	// Digits after point
	while (szData != szEnd)
	{
		if (!(*szData >= '0' && *szData <= '9')) { break; }
		++ szData;
	}

	// Scientific?
	if (szData == szEnd || (*szData != 'E' && *szData != 'e'))
	{
		INT_32 iDigitSize = szEnd() - szStart();
		CHAR_P szDigit = new CHAR_8[iDigitSize + 1];
		memcpy(szDigit, szStart(), iDigitSize);
		szDigit[iDigitSize] = '\0';

		double iFF = 0;
		sscanf(szDigit, "%lg", &iFF);

		dFloatData = iFF;
		iParsedNumberType = 1;

		delete [] szDigit;
		return szData;
	}
	else
	{
		INT_32 iFoundDigits = 0;
		++ szData;
		// Error?
		if (szData == szEnd) { throw CTPPParserSyntaxError("exponent has no digits", szData.GetLine(), szData.GetLinePos()); }
		// Exponent sign
		if (*szData == '-' || *szData == '+') { ++ szData; }
		// Error?
		if (szData == szEnd) { throw CTPPParserSyntaxError("exponent has no digits", szData.GetLine(), szData.GetLinePos()); }
		// Exponent value
 		while (szData != szEnd)
		{
			if (!(*szData >= '0' && *szData <= '9')) { break; }
			iFoundDigits = 1;
			++ szData;
		}
		if (iFoundDigits == 0) { throw CTPPParserSyntaxError("exponent has no digits", szData.GetLine(), szData.GetLinePos()); }

//fprintf(stderr, "start = %s", szStart());

		INT_32 iDigitSize = szEnd() - szStart();
		CHAR_P szDigit = new CHAR_8[iDigitSize + 1];
		memcpy(szDigit, szStart(), iDigitSize);
		szDigit[iDigitSize] = '\0';

		double iFF = 0;
		sscanf(szDigit, "%le", &iFF);

		dFloatData = iFF;
		iParsedNumberType = 1;

		delete [] szDigit;
		return szData;
	}

return NULL;
}

//
// CTPP2 function [a-zA-Z_][a-zA-Z0-9_]( Expr1 [, Expr2 ... ] )
//
CCharIterator CTPP2Parser::IsFunc(CCharIterator szData, CCharIterator szEnd, CCharIterator & sFuncEnd, UINT_32 & iFunctionParams)
{
	eCTPP2ExprOperator eResultOperator = EXPR_UNDEF;
	CCharIterator sTMP = IsVar(szData, szEnd);

	if (sTMP != NULL)
	{

		sFuncEnd = sTMP;
		szData = sTMP;
		// Skip white space
		sTMP = IsWhiteSpace(szData, szEnd, 0);

		// '(' blah, clah, mlah ')'
		// Not a function; perhaps, variable, value, or string constant
		if (sTMP == NULL || *sTMP != '(') { return NULL; }

		// OK
		++sTMP;
		if (sTMP == szEnd) { throw CTPPParserSyntaxError("expected primary-expression", sTMP.GetLine(), sTMP.GetLinePos()); }

		// ... blah, clah, mlah ')'
		if (*sTMP == ')') { return ++sTMP; }

		for(;;)
		{
			szData = sTMP;
			// Expression
			sTMP = IsExpr(szData, szEnd, eResultOperator);
			if (sTMP == NULL) { throw "Ouch!"; }

			if (eResultOperator == EXPR_FUNCTION)
			{
				pCTPP2Compiler -> OutputFnResult();
			}

			++iFunctionParams;
			// Skip white space
			sTMP = IsWhiteSpace(sTMP, szEnd, 0);

			// , clah, mlah ')'
			if      (*sTMP == ',')
			{
				++sTMP;
			}
			else if (*sTMP == ')') { ++sTMP; break; }
		}
		return sTMP;
	}
//fprintf(stderr, "IsFunc: ----------------------------------> NULL\n");
return NULL;
}

//
// CTPP2 white space ('\t' | ' ')
//
CCharIterator CTPP2Parser::IsWhiteSpace(CCharIterator szData, CCharIterator szEnd, const UINT_32 & iSkipFirst)
{
	if (iSkipFirst == 1)
	{
		if (szData != szEnd && (*szData != ' ' && *szData != '\t')) { return NULL; }
		++szData;
	}

	while (szData != szEnd)
	{
		if (*szData != ' ' && *szData != '\t') { break; }
		++szData;
	}
return szData;
}

//
// CTPP2 variable <TMPL_var Expr>
//
CCharIterator CTPP2Parser::VarOperator(CCharIterator szData, CCharIterator szEnd)
{
	eCTPP2ExprOperator eResultOperator;
	// Skip white space
	CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
	szData = sTMP;

	sTMP = IsExpr(szData, szEnd, eResultOperator);
	if (sTMP == NULL)
	{
		throw "Ouch!";
	}
	szData = sTMP;

	// Skip white space
	szData = IsWhiteSpace(szData, szEnd, 0);

	if (*szData != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", szData.GetLine(), szData.GetLinePos()); }

	if (eResultOperator == EXPR_FUNCTION)
	{
		pCTPP2Compiler -> OutputFnResult();
	}

	// Output variable
	pCTPP2Compiler -> OutputVariable();

return szData;
}

//
// Store comparison result (if branch)
//
UINT_32 CTPP2Parser::StoreIfComparisonResult(CCharIterator szData, const eCTPP2ExprOperator & eResultOperator)
{
	UINT_32 iElseBranchCallerIP = (UINT_32)-1;

	// Compare values
	switch (eResultOperator)
	{
		// Simple case: comparison result is always true or always false
		// 0 ?
		case EXPR_INT_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of INTEGER VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (iIntData != 0) { fprintf(stderr, "true\n"); }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "false\n");
			}

			break;

		// 0.0 ?
		case EXPR_FLOAT_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of FLOAT VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (dFloatData != 0.0) { fprintf(stderr, "true\n");  }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "true\n");
			}
			break;

		// Empty string?
		case EXPR_STRING_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of STRING VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (sTMPBuf.size() != 0) { fprintf(stderr, "true\n");  }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "true\n");
			}
			break;

		// 1-st complex case: compare variable
		case EXPR_VARIABLE:
			pCTPP2Compiler -> ExistStackVariable();
			pCTPP2Compiler -> PopVariable();
			iElseBranchCallerIP = pCTPP2Compiler -> NEJump(UINT_32(-1));
			break;

		// 2-nd complex case: compare function execution result
		case EXPR_FUNCTION:
			pCTPP2Compiler -> ExistARReg();
			iElseBranchCallerIP = pCTPP2Compiler -> NEJump(UINT_32(-1));
			break;

		default:
			throw "Ouch!";
	}

return iElseBranchCallerIP;
}

//
// Store comparison result (unless branch)
//
UINT_32 CTPP2Parser::StoreUnlessComparisonResult(CCharIterator szData, const eCTPP2ExprOperator & eResultOperator)
{
	UINT_32 iElseBranchCallerIP = (UINT_32)-1;

	// Compare values
	switch (eResultOperator)
	{
		// Simple case: comparison result is always true or always false
		// 0 ?
		case EXPR_INT_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of INTEGER VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (iIntData == 0) { fprintf(stderr, "true\n"); }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "false\n");
			}
			break;

		// 0.0 ?
		case EXPR_FLOAT_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of FLOAT VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (dFloatData == 0.0) { fprintf(stderr, "true\n");  }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "true\n");
			}
			break;

		// Empty string?
		case EXPR_STRING_VALUE:
			fprintf(stderr, "WARNING: near line %d, pos. %d: comparison result of STRING VALUE is always ", szData.GetLine(),  szData.GetLinePos());
			pCTPP2Compiler -> RemoveInstr();

			if (sTMPBuf.size() == 0) { fprintf(stderr, "true\n");  }
			else
			{
				// Store point to unconditional jump (to ELSE branch)
				iElseBranchCallerIP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
				fprintf(stderr, "true\n");
			}
			break;

		// 1-st complex case: compare variable
		case EXPR_VARIABLE:
			pCTPP2Compiler -> ExistStackVariable();
			pCTPP2Compiler -> PopVariable();
			iElseBranchCallerIP = pCTPP2Compiler -> EQJump(UINT_32(-1));
			break;

		// 2-nd complex case: compare function execution result
		case EXPR_FUNCTION:
			pCTPP2Compiler -> ExistARReg();
			iElseBranchCallerIP = pCTPP2Compiler -> EQJump(UINT_32(-1));
			break;

		default:
			throw "Ouch!";
	}

return iElseBranchCallerIP;
}

//
//  <TMPL_if IfExpr> blah-blah-blah [ <TMPL_else> clah-clah-clah ] </TMPL_if>
//
CCharIterator CTPP2Parser::IfOperator(CCharIterator szData, CCharIterator szEnd)
{
	std::vector<INT_32> vJMPPoints;

	eCTPP2ExprOperator eResultOperator = EXPR_UNDEF;
	// Skip white space
	CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
	szData = sTMP;

	// Condition
	szData = IsIfExpr(szData, szEnd, eResultOperator);

	// Syntax ok, store result
	INT_32 iBranchCallerIP = StoreIfComparisonResult(szData, eResultOperator);

	eBreakFound = TMPL_if;
//fprintf(stderr, "3IfOperator--> `%s`\n", szData());

	// Recursive descend
	sTMP = Parse(++szData, szEnd);
	if (sTMP == NULL)
	{
		throw "Ouch!";
	}
	szData = sTMP;

//fprintf(stderr, "4IfOperator--> `%s`\n", szData());
	if (eBreakFound != TMPL_else  &&
	    eBreakFound != TMPL_elsif &&
	    eBreakFound != TMPL_if) { throw CTPPParserOperatorsMismatch("<TMPL_else> OR <TMPL_elsif expression> OR </TMPL_if>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

	// Jump to end of branch
	UINT_32 iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
	vJMPPoints.push_back(iTMP);
//fprintf(stderr, "if -----------> iTMP `%d`\n", iTMP);

	while (eBreakFound == TMPL_elsif)
	{
		eResultOperator = EXPR_UNDEF;

		// Skip white space
		CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
		if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
		szData = sTMP;

		INT_32 iBranchEntryPoint = pCTPP2Compiler -> GetCodeSize();

		// Condition
		szData = IsIfExpr(szData, szEnd, eResultOperator);

		// Syntax ok, store result
		INT_32 iBranchCodePoint = StoreIfComparisonResult(szData, eResultOperator);

		if (iBranchCallerIP != -1)
		{
			// Set JXX point for previous branch
			VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(iBranchCallerIP);
			pInstruction -> argument = iBranchEntryPoint;
		}
		iBranchCallerIP = iBranchCodePoint;

		eBreakFound = TMPL_elsif;
//fprintf(stderr, "5IfOperator--> `%s`\n", szData());

		// Recursive descend
		sTMP = Parse(++szData, szEnd);
		if (sTMP == NULL)
		{
			throw "Ouch!";
		}

		if (eBreakFound != TMPL_else  &&
		    eBreakFound != TMPL_elsif &&
		    eBreakFound != TMPL_if) { throw CTPPParserOperatorsMismatch("<TMPL_else> OR <TMPL_elsif expression> OR </TMPL_if>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

		szData = sTMP;

		// Jump to end of branch
		iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
		vJMPPoints.push_back(iTMP);
//fprintf(stderr, "elsif -----------> iTMP `0x%08X`\n", iTMP);
	}

	// Jump to else-branch if condition is false
	INT_32 iExitPoint = pCTPP2Compiler -> GetCodeSize();
	if (iBranchCallerIP != -1)
	{
		VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(iBranchCallerIP);
		pInstruction -> argument = iExitPoint;
	}

	// Else branch
	if (eBreakFound == TMPL_else)
	{
		sTMP = Parse(++szData, szEnd);
		if (sTMP == NULL)
		{
			throw "Ouch!";
		}
		szData = sTMP;
		// Jump to end of branch
		iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
		vJMPPoints.push_back(iTMP);
		iExitPoint = pCTPP2Compiler -> GetCodeSize();
	}

	// /TMPL_if
	if (eBreakFound != TMPL_if) { throw CTPPParserOperatorsMismatch("</TMPL_if>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

	// Remove last JMP, if need
	INT_32 iLastJMPPos = vJMPPoints.size() - 1;

	if (iLastJMPPos > 0 && iExitPoint - 1 == vJMPPoints[iLastJMPPos])
	{
		pCTPP2Compiler -> RemoveInstr();
		-- iExitPoint;
		-- iLastJMPPos;
	}

	// Okay, set JMP points
	for(INT_32 iI = 0; iI <= iLastJMPPos; ++iI)
	{
		VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(vJMPPoints[iI]);
		pInstruction -> argument = iExitPoint;
	}

return szData;
}

//
//  <TMPL_unless IfExpr> blah-blah-blah [ <TMPL_else> clah-clah-clah ] </TMPL_unless>
//
CCharIterator CTPP2Parser::UnlessOperator(CCharIterator szData, CCharIterator szEnd)
{
	std::vector<INT_32> vJMPPoints;

	eCTPP2ExprOperator eResultOperator = EXPR_UNDEF;
	// Skip white space
	CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
	szData = sTMP;

	// Condition
	szData = IsIfExpr(szData, szEnd, eResultOperator);

	// Syntax ok, store result
	INT_32 iBranchCallerIP = StoreUnlessComparisonResult(szData, eResultOperator);

	eBreakFound = TMPL_if;
//fprintf(stderr, "3IfOperator--> `%s`\n", szData());

	// Recursive descend
	sTMP = Parse(++szData, szEnd);
	if (sTMP == NULL)
	{
		throw "Ouch!";
	}
	szData = sTMP;

//fprintf(stderr, "4IfOperator--> `%s`\n", szData());
	if (eBreakFound != TMPL_else  &&
	    eBreakFound != TMPL_elsif &&
	    eBreakFound != TMPL_unless) { throw CTPPParserOperatorsMismatch("<TMPL_else> OR <TMPL_elsif expression> OR </TMPL_if>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

	// Jump to end of branch
	UINT_32 iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
	vJMPPoints.push_back(iTMP);
//fprintf(stderr, "if -----------> iTMP `%d`\n", iTMP);

	while (eBreakFound == TMPL_elsif)
	{
		eResultOperator = EXPR_UNDEF;

		// Skip white space
		CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
		if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
		szData = sTMP;

		INT_32 iBranchEntryPoint = pCTPP2Compiler -> GetCodeSize();

		// Condition
		szData = IsIfExpr(szData, szEnd, eResultOperator);

		// Syntax ok, store result
		INT_32 iBranchCodePoint = StoreIfComparisonResult(szData, eResultOperator);

		// Set JXX point for previous branch
		if (iBranchCallerIP != -1)
		{
			VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(iBranchCallerIP);
			pInstruction -> argument = iBranchEntryPoint;
		}

		iBranchCallerIP = iBranchCodePoint;

		eBreakFound = TMPL_elsif;
//fprintf(stderr, "5IfOperator--> `%s`\n", szData());

		// Recursive descend
		sTMP = Parse(++szData, szEnd);
		if (sTMP == NULL)
		{
			throw "Ouch!";
		}

		if (eBreakFound != TMPL_else  &&
		    eBreakFound != TMPL_elsif &&
		    eBreakFound != TMPL_unless) { throw CTPPParserOperatorsMismatch("<TMPL_else> OR <TMPL_elsif expression> OR </TMPL_unless>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

		szData = sTMP;

		// Jump to end of branch
		iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
		vJMPPoints.push_back(iTMP);
//fprintf(stderr, "elsif -----------> iTMP `0x%08X`\n", iTMP);
	}

	// Jump to else-branch if condition is false
	INT_32 iExitPoint = pCTPP2Compiler -> GetCodeSize();
	if (iBranchCallerIP != -1)
	{
		VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(iBranchCallerIP);
		pInstruction -> argument = iExitPoint;
	}

	// Else branch
	if (eBreakFound == TMPL_else)
	{
		sTMP = Parse(++szData, szEnd);
		if (sTMP == NULL)
		{
			throw "Ouch!";
		}
		szData = sTMP;
		// Jump to end of branch
		iTMP = pCTPP2Compiler -> UncondJump(UINT_32(-1));
		vJMPPoints.push_back(iTMP);
		iExitPoint = pCTPP2Compiler -> GetCodeSize();
	}

	// /TMPL_if
	if (eBreakFound != TMPL_unless) { throw CTPPParserOperatorsMismatch("</TMPL_unless>", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

	// Remove last JMP, if need
	INT_32 iLastJMPPos = vJMPPoints.size() - 1;

	if (iLastJMPPos > 0 && iExitPoint - 1 == vJMPPoints[iLastJMPPos])
	{
		pCTPP2Compiler -> RemoveInstr();
		-- iExitPoint;
		-- iLastJMPPos;
	}

	// Okay, set JMP points
	for(INT_32 iI = 0; iI <= iLastJMPPos; ++iI)
	{
		VMInstruction * pInstruction = pCTPP2Compiler -> GetInstruction(vJMPPoints[iI]);
		pInstruction -> argument = iExitPoint;
	}

return szData;
}

//
// CTPP2 loop <TMPL_loop Expr> blah-blah-blah </TMPL_loop>
//
CCharIterator CTPP2Parser::LoopOperator(CCharIterator szData, CCharIterator szEnd)
{
	CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
	szData = sTMP;

	UINT_32 iRetPoint = 0;
	// Parse loop expression
	szData = IsLoopExpr(szData, szEnd, iRetPoint);

	// Skip white space
	szData = IsWhiteSpace(szData, szEnd, 0);

	if (*szData != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", szData.GetLine(), szData.GetLinePos()); }

	// Recursive descend
	sTMP = Parse(++szData, szEnd);
	if (sTMP == NULL)
	{
		throw "Ouch!";
	}

	if (eBreakFound != TMPL_loop ) { throw CTPPParserOperatorsMismatch("</TMPL_loop>", GetOperatorName(eBreakFound), sTMP.GetLine(), sTMP.GetLinePos()); }

	// Store point
	pCTPP2Compiler -> ResetScope(iRetPoint);

return sTMP;
}

//
// CTPP2 include <TMPL_include Expr>
//
CCharIterator CTPP2Parser::IncludeOperator(CCharIterator szData, CCharIterator szEnd)
{
	CCharIterator sTMP = IsWhiteSpace(szData, szEnd);
	if (sTMP == NULL) { throw CTPPParserSyntaxError("expected at least one space symbol", szData.GetLine(), szData.GetLinePos()); }
	szData = sTMP;

	// Check include file name
	sTMP = IsString(szData, szEnd);
	if (sTMP == NULL)
	{
		// Check old-fashion syntax
		sTMP = IsStringCompatOldVersion(szData, szEnd);

		if (sTMP == NULL) { throw CTPPParserSyntaxError("incorrect include file name", szData.GetLine(), szData.GetLinePos()); }
	}

	// Skip white space
	sTMP = IsWhiteSpace(sTMP, szEnd, 0);

	// '>'
	if (*sTMP != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", sTMP.GetLine(), sTMP.GetLinePos()); }

	// All okay, need to parse include file
	CTPP2SourceLoader * pTMPSourceLoader = NULL;
	try
	{
		// Create loader
		pTMPSourceLoader = pSourceLoader -> Clone();
		// Load template
		pTMPSourceLoader -> LoadTemplate(sTMPBuf.c_str());
		// Create parser
		CTPP2Parser oTMPParser(pTMPSourceLoader, pCTPP2Compiler);
		// No HLT at end of code
		oTMPParser.Compile(0);
	}
	catch(CTPPLogicError & e)
	{
		delete pTMPSourceLoader;

		CHAR_8 szTMPBuffer[2048 + 1];
		snprintf(szTMPBuffer, 2048, "Cannot load template file `%s`: %s", sTMPBuf.c_str(), e.what());

		throw CTPPParserSyntaxError(szTMPBuffer, sTMP.GetLine(), sTMP.GetLinePos());
	}
	catch(CTPPUnixException & e)
	{
		delete pTMPSourceLoader;

		CHAR_8 szTMPBuffer[2048 + 1];
		snprintf(szTMPBuffer, 2048, "Cannot load template file `%s`: %s", sTMPBuf.c_str(), strerror(e.ErrNo()));

		throw CTPPParserSyntaxError(szTMPBuffer, sTMP.GetLine(), sTMP.GetLinePos());
	}
	catch(CTPPParserException & e)
	{
		delete pTMPSourceLoader;

		CHAR_8 szTMPBuffer[2048 + 1];
		snprintf(szTMPBuffer, 2048, "In include file '%s' at line %d, pos %d: %s", sTMPBuf.c_str(), e.GetLine(), e.GetLinePos(), e.what());

		throw CTPPParserSyntaxError(szTMPBuffer, sTMP.GetLine(), sTMP.GetLinePos());
	}

	delete pTMPSourceLoader;

return sTMP;
}

//
// CTPP2 comment &ltTMPL_comment&gt; ..... &lt/TMPL_comment&gt;
//
CCharIterator CTPP2Parser::CommentOperator(CCharIterator szData, CCharIterator szEnd)
{
	// Skip white space
	szData = IsWhiteSpace(szData, szEnd, 0);

	// '>'
	if (*szData != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", szData.GetLine(), szData.GetLinePos()); }

	++szData;

	// Skip code
	for (;;)
	{
		// Open symbol
		while (szData != szEnd && *szData != TMPL_OPEN_SYMBOL) { ++szData; }
		// Unexpected end of template
		if (szData == szEnd) { throw CTPPParserSyntaxError("expected '</TMPL_comment>'", szData.GetLine(), szData.GetLinePos()); }

		CCharIterator szTMP = IsCloseTag(++szData, szEnd);
		if (szTMP == NULL) { ++szData; }
		else
		{
			eCTPP2Operator oOperatorType = UNDEF;
			szData = szTMP;
			szTMP = IsOperator(++szData, szEnd, oOperatorType);
			//
			if (szTMP == szEnd) { throw CTPPParserSyntaxError("unexpected end of file", szData.GetLine(), szData.GetLinePos()); }
			// Operator NOT found after opener tag
			if (szTMP != NULL)
			{
				if (oOperatorType == TMPL_comment)
				{
					szTMP = IsWhiteSpace(szTMP, szEnd, 0);
					if (*szTMP != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", szTMP.GetLine(), szTMP.GetLinePos()); }
					return szTMP;
				}
				szData = szTMP;
			}

		}
	}
return NULL;
}

//
// Parse CTPP2 source code
//
CCharIterator CTPP2Parser::Parse(CCharIterator szData, CCharIterator szEnd)
{
	eBreakFound = UNDEF;

	CCharIterator szIter = szData;
	CCharIterator szText = szData;

	for (;;)
	{
		// Open tag
		while (szIter != szEnd && *szIter != TMPL_OPEN_SYMBOL) { ++szIter; }
		if (szIter == szEnd)
		{
			// Store static data
			pCTPP2Compiler -> OutputStaticData(szText(), szEnd() - szText());

			return szEnd;
		}

		// Check open tag
		CCharIterator sTMP = IsOpenTag(++szIter, szEnd);

		if (sTMP != NULL)
		{
			// Store static data
			pCTPP2Compiler -> OutputStaticData(szText(), szIter() - szText() - 1);

			eCTPP2Operator oOperatorType = UNDEF;
			//fprintf(stderr, "`%s`\n", sTMP);
			szIter = sTMP;
			// Check operator
			sTMP = IsOperator(++szIter, szEnd, oOperatorType);

			// Operator NOT found after opener tag
			if (sTMP == NULL)  { throw CTPPParserSyntaxError("incorrect operator", szIter.GetLine(), szIter.GetLinePos()); }

			if (sTMP == szEnd) { throw CTPPParserSyntaxError("unexpected end of file", szIter.GetLine(), szIter.GetLinePos()); }
			// Operator found
			szIter = sTMP;

			switch (oOperatorType)
			{
				// Print variable or execute function
				// User defined function
				case TMPL_udf:
//					fprintf(stderr, "Synonym for TMPL_var\n");

				// Variable
				case TMPL_var:
//					fprintf(stderr, "TMPL_var\n");
					sTMP = VarOperator(szIter, szEnd);
					break;

				// Branch
				case TMPL_if:
//					fprintf(stderr, "TMPL_if\n");
					sTMP = IfOperator(szIter, szEnd);
					break;

				// Next branch
				case TMPL_elsif:
//					fprintf(stderr, "TMPL_elsif\n");
					eBreakFound = TMPL_elsif;
					return sTMP;
					break;

				// Branch with inverted condition
				case TMPL_unless:
//					fprintf(stderr, "TMPL_unless\n");
					sTMP = UnlessOperator(szIter, szEnd);
					break;

				// Else-branch
				case TMPL_else:
//					fprintf(stderr, "TMPL_else\n");
					eBreakFound = TMPL_else;
					return szIter;

				// Loop
				case TMPL_loop:
//					fprintf(stderr, "TMPL_loop\n");
					eBreakFound = TMPL_loop;
					sTMP = LoopOperator(szIter, szEnd);
					break;

				// Include file
				case TMPL_include:
//					fprintf(stderr, "TMPL_include\n");
					sTMP = IncludeOperator(szIter, szEnd);
					break;

				// Comment
				case TMPL_comment:
//					fprintf(stderr, "TMPL_comment\n");
					sTMP = CommentOperator(szIter, szEnd);
					break;

				// Error!
				default:
					throw CTPPParserSyntaxError("impossible happened", szIter.GetLine(), szIter.GetLinePos());
					break;
			}
			szIter = sTMP;
			++ szIter;
			szText = szIter;
		}
		else
		{
			sTMP = IsCloseTag(szIter, szEnd);
			if (sTMP != NULL)
			{
				// Store static data
				pCTPP2Compiler -> OutputStaticData(szText(), szIter() - szText() - 1);

				eCTPP2Operator oOperatorType = UNDEF;
				//fprintf(stderr, "`%s`\n", sTMP);
				szIter = sTMP;
				// Check operator
				sTMP = IsOperator(++szIter, szEnd, oOperatorType);

				switch (oOperatorType)
				{
					// Error!
					case UNDEF:
						throw CTPPParserSyntaxError("impossible happened", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_udf:
						throw CTPPParserSyntaxError("operator '</TMPL_udf>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_var:
						throw CTPPParserSyntaxError("operator '</TMPL_var>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_else:
						throw CTPPParserSyntaxError("operator '</TMPL_else>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_elsif:
						throw CTPPParserSyntaxError("operator '</TMPL_elsif>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_include:
						throw CTPPParserSyntaxError("operator '</TMPL_include>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_break:
						throw CTPPParserSyntaxError("operator '</TMPL_break>' does not exist", szIter.GetLine(), szIter.GetLinePos());

					case TMPL_comment:
						throw CTPPParserSyntaxError("operator '</TMPL_comment>' cannot use heret", szIter.GetLine(), szIter.GetLinePos());

					// Branch
					case TMPL_if:
//						fprintf(stderr, "/TMPL_if\n");

						sTMP =  IsWhiteSpace(sTMP, szEnd, 0);
						if (*sTMP != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", sTMP.GetLine(), sTMP.GetLinePos()); }
						eBreakFound = TMPL_if;
						return sTMP;

					// Branch with inverted condition
					case TMPL_unless:
//						fprintf(stderr, "/TMPL_unless\n");

						sTMP =  IsWhiteSpace(sTMP, szEnd, 0);
						if (*sTMP != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", sTMP.GetLine(), sTMP.GetLinePos()); }
						eBreakFound = TMPL_unless;
						return sTMP;

					// Loop
					case TMPL_loop:
//						fprintf(stderr, "/TMPL_loop\n");

						sTMP =  IsWhiteSpace(sTMP, szEnd, 0);
						if (*sTMP != TMPL_CLOSE_SYMBOL) { throw CTPPParserSyntaxError("expected '>'", sTMP.GetLine(), sTMP.GetLinePos()); }
						eBreakFound = TMPL_loop;
						return sTMP;

					default:
						throw "UndefOperator";
				}

				++ szIter;
				szText = szIter;
			}
			else
			{
				if (szIter == szEnd)
				{
					// Store static data
					pCTPP2Compiler -> OutputStaticData(szText(), szEnd() - szText());

					break;
				}

				++ szIter;
			}
		}
	}

// This should not happened
return NULL;
}

//
// Get printable operator name
//
CCHAR_P CTPP2Parser::GetOperatorName(const eCTPP2Operator & oOperator)
{
	switch(oOperator)
	{
		case TMPL_var:     return "TMPL_var";
		case TMPL_if:      return "TMPL_if";
		case TMPL_unless:  return "TMPL_unless";
		case TMPL_else:    return "TMPL_else";
		case TMPL_elsif:   return "TMPL_elsif";
		case TMPL_loop:    return "TMPL_loop";
		case TMPL_include: return "TMPL_include";
		case TMPL_udf:     return "TMPL_udf";
		case TMPL_comment: return "TMPL_comment";
		case TMPL_break:   return "TMPL_break";
		default:
			 return "*UNDEF*";
	}
// Make stupid compiler happy
return NULL;
}

//
// Compile CTPP2 template
//
INT_32 CTPP2Parser::Compile(const UINT_32 & iHalt)
{
	UINT_32 iTemplateSize = 0;
	CCharIterator szData = pSourceLoader -> GetTemplate(iTemplateSize);

	// Impossible happened.
	if (szData == NULL) { return -1; }

	CCharIterator szEnd = szData() + iTemplateSize;
	CCharIterator sTMP = Parse(szData, szEnd);

	if (sTMP != szEnd)
	{
		if (eBreakFound != UNDEF) { throw CTPPParserOperatorsMismatch("absence", GetOperatorName(eBreakFound), szData.GetLine(), szData.GetLinePos()); }

		throw CTPPParserSyntaxError("syntax error", sTMP.GetLine(), sTMP.GetLinePos());
	}

	if (iHalt == 1)
	{
		// Halt program
		pCTPP2Compiler -> Halt();
	}

return 0;
}

//
// A destructor
//
CTPP2Parser::~CTPP2Parser() throw()
{
	;;
}

} // namespace CTPP
// End.
