/*-
 * Copyright (c) 2006, 2007 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      VM.cpp
 *
 * $CTPP$
 */

#include <VM.hpp>
#include <VMOpcodes.h>
#include <VMException.hpp>
#include <VMStackException.hpp>
#include <VMInstruction.hpp>

#include <string.h>

/*
    Foreground colors
       30    Black
       31    Red
       32    Green
       33    Yellow
       34    Blue
       35    Magenta
       36    Cyan
       37    White

    Background colors
       40    Black
       41    Red
       42    Green
       43    Yellow
       44    Blue
       45    Magenta
       46    Cyan
       47    White
*/
#define BRIGHT 1
#define RED    31
#define YELLOW 33
#define GREEN  32
#define BLUE   34
#define BG_BLACK 40


#define HL_CODE(x) fprintf(stderr, "%c[%d;%d;%dm", 0x1B, BRIGHT, (x), BG_BLACK)
#define HL_RST  fprintf(stderr, "%c[%dm", 0x1B, 0)


namespace CTPP // C++ Template Engine
{

//
// Constructor
//
VM::VM(SyscallFactory  & oISyscallFactory,
       const UINT_32   & iIMaxArgStackSize,
       const UINT_32   & iIMaxCodeStackSize,
       const UINT_32   & iIMaxSteps,
       const UINT_32   & iIDebugLevel): oSyscallFactory(oISyscallFactory),
                                        iMaxArgStackSize(iIMaxArgStackSize),
                                        iMaxCodeStackSize(iIMaxCodeStackSize),
                                        iMaxSteps(iIMaxSteps),
                                        iDebugLevel(iIDebugLevel),
                                        iMaxCalls(0),
                                        iMaxUsedCalls(0),
                                        aCallTranslationMap(NULL),
                                        oVMArgStack(iMaxArgStackSize),
                                        oVMCodeStack(iMaxCodeStackSize)
{
	;;
}

//
// Initialize virtual machine
//
INT_32 VM::Init(OutputCollector     & pOutputCollector,
                const VMMemoryCore  & oMemoryCore)
{
	// Create syscalls translation map
	iMaxUsedCalls = oMemoryCore.syscalls.GetRecordsNum();
	// Cache
	if (iMaxUsedCalls >= iMaxCalls)
	{
		delete [] aCallTranslationMap;
		aCallTranslationMap = new SyscallHandler *[iMaxUsedCalls];
		iMaxCalls = iMaxUsedCalls;
	}
	// Initialize program
	for (UINT_32 iCallNum = 0; iCallNum < iMaxUsedCalls; ++iCallNum)
	{
		// Ugh
		UINT_32 iCallNameLength = 0;
		// NULL-terminated string, iCallNameLength will not used
		CCHAR_P sCallName = oMemoryCore.syscalls.GetData(iCallNum, iCallNameLength);

		// Get syscall
		SyscallHandler * pTMP = oSyscallFactory.GetHandlerByName(sCallName);

		// If this syscall is not present in factory, throw exception
		if (pTMP == NULL) { throw InvalidSyscall(sCallName, 0); }

		// All OK
		aCallTranslationMap[iCallNum] = pTMP;

		// Initialize syscall handler
		pTMP -> PreExecuteSetup(pOutputCollector,
		                        DR,
		                        oMemoryCore.syscalls,
		                        oMemoryCore.static_data,
		                        oMemoryCore.static_text);
	}

return 0;
}

//
// Run program
//
INT_32 VM::Run(const VMMemoryCore  & oMemoryCore,
               UINT_32             & iIP,
               CDT                 & oCDT)
{
	DR = oCDT;
	// Get code segment
	const VMInstruction * aCode = oMemoryCore.instructions;
	const UINT_32 iCodeLength   = oMemoryCore.code_size;
	UINT_32 iExecutedSteps      = 0;

	try
	{
		while (iIP < iCodeLength)
		{
			UINT_32 iOpCode = aCode[iIP].instruction;

#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "CODE 0x%08X ARG 0x%08X | ", iOpCode, aCode[iIP].argument);
HL_RST;
#endif

			switch(iOpCode)
			{
				// Halt program
				case HLT:
//				if      (iOpCode == HLT)
				{
#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "0x%08X HLT\n", iIP);
HL_RST;
#endif
					return 0;
				}
				break;

				// Break execution
				case BRK:
//				else if (iOpCode == BRK)
				{
#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "0x%08X BRK\n", iIP);
HL_RST;
#endif
					++iIP;
					if (iDebugLevel > 0) { return 0; }
				}
				break;

				// System call
				case SYSCALL:
//				else if (iOpCode == SYSCALL)
				{
					const UINT_32 iCallNum    = (aCode[iIP].argument & 0xFFFF0000) >> 16;
					const UINT_32 iCallArgNum = (aCode[iIP].argument & 0x0000FFFF);

#ifdef _DEBUG
{
UINT_32 iCallNameLength = 0;
// NULL-terminated string, iCallNameLength will not used
CCHAR_P sCallName = oMemoryCore.syscalls.GetData(iCallNum, iCallNameLength);
HL_CODE(YELLOW);
fprintf(stderr, "0x%08X SYSCALL   0x%08X %s(ARGS: %d)\n", iIP, iCallNum, sCallName, iCallArgNum);
HL_RST;
}
#endif
					// Check call number
					if (iCallNum > iMaxUsedCalls) { throw InvalidSyscall("*** CORRUPTED ***", iIP); }

					// Invoke handler
					if (aCallTranslationMap[iCallNum] -> Handler(oVMArgStack.GetStackFrame(), iCallArgNum, oRegs[0]) != 0)
					{
						throw InvalidSyscall("*** Internal syscall error ***", iIP);
					}

					// Clear stack
					oVMArgStack.ClearStack(iCallArgNum);

					++iIP;
				}
				break;

				// Call procedure
				case CALL:
//				else if (iOpCode == CALL)
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X CALL      0x%08X\n", iIP, aCode[iIP].argument);
#endif
					if (iExecutedSteps >= iMaxSteps) { throw ExecutionLimitReached(iIP); }

					const UINT_32 iNewIP = aCode[iIP].argument;
					if (iNewIP >= iCodeLength) { throw CodeSegmentOverrun(iIP); }

					oVMCodeStack.PushAddress(iIP + 1);
					iIP = aCode[iIP].argument;
				}
				break;

				// Return from procedure
				case RET:
//				else if (iOpCode == RET)
				{

#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X RET       %d\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					if (iExecutedSteps >= iMaxSteps) { throw ExecutionLimitReached(iIP); }
					// Clear stack
					oVMArgStack.ClearStack(aCode[iIP].argument);
					// Return
					iIP = oVMCodeStack.PopAddress();
				}
				break;

				// Unconditional jump
				case JMP:
//			else if (iOpCode == JMP)
				{
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X JMP       0x%08X\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					if (iExecutedSteps >= iMaxSteps) { throw ExecutionLimitReached(iIP); }

					const UINT_32 iNewIP = aCode[iIP].argument;
					if (iNewIP >= iCodeLength) { throw CodeSegmentOverrun(iIP); }

					iIP = iNewIP;
				}
				break;

				// Pop value from stack
				case POP:
//				else if (iOpCode == POP)
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X POP       %d\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					oVMArgStack.ClearStack(aCode[iIP].argument);
					++iIP;
				}
				break;

				case NOP:
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X NOP\n", iIP);
HL_RST;
#endif
					++iIP;
				}
				break;

				// All other operations
				default:
//			else
				{
					UINT_32 iHiOpcode   = iOpCode & 0xFFFF0000;
					UINT_32 iLoHiOpcode = iOpCode & 0x0000FF00;
					UINT_32 iLoLoOpcode = iOpCode & 0x000000FF;

					if (iHiOpcode == ADD)
					{
						CR = AR + BR;
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X ADD       ('%s', '%s') -> '%s'\n", iIP, AR.GetString().c_str(), BR.GetString().c_str(), CR.GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					else if (iHiOpcode == SUB)
					{
						CR = AR - BR;
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X SUB       ('%s', '%s') -> '%s'\n", iIP, AR.GetString().c_str(), BR.GetString().c_str(), CR.GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					else if (iHiOpcode == MUL)
					{
						CR = AR * BR;
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MUL       ('%s', '%s') -> '%s'\n", iIP, AR.GetString().c_str(), BR.GetString().c_str(), CR.GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					else if (iHiOpcode == DIV)
					{
						CR = AR / BR;
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X DIV       ('%s', '%s') -> '%s'\n", iIP, AR.GetString().c_str(), BR.GetString().c_str(), CR.GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					// Increment register
					else if (iHiOpcode == INC)
					{
						if (iLoLoOpcode <= ARG_SRC_LASTREG)
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X INC       %cR (%d)\n", iIP, CHAR_8(iLoLoOpcode + 'A'), INT_32(oRegs[iLoLoOpcode].GetInt()));
HL_RST;
#endif
							++oRegs[iLoLoOpcode];
						}
						else { throw IllegalOpcode(iIP, iOpCode); }

						++iIP;
					}
					// Decrement register
					else if (iHiOpcode == DEC)
					{
						if (iLoLoOpcode <= ARG_SRC_LASTREG)
						{

#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X DEC       %cR (%d)\n",  iIP, CHAR_8(iLoLoOpcode + 'A'), INT_32(oRegs[iLoLoOpcode].GetInt()));
HL_RST;
#endif
							--oRegs[iLoLoOpcode];
						}
						else { throw IllegalOpcode(iIP, iOpCode); }

						++iIP;
					}
					// Compare two registers: AR and BR
					else if (iHiOpcode == CMP)
					{
						if (iLoHiOpcode <= ARG_DST_LASTREG)
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							// Register-to-register
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{

#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X CMP       %cR, %cR", iIP, CHAR_8(iLoLoOpcode + 'A'), CHAR_8(iDstReg + 'A'));
HL_RST;
#endif
								CDT & oFirst  = oRegs[iLoLoOpcode];
								CDT & oSecond = oRegs[iDstReg];

								// Compare as values
								if (oFirst.GetType() <= CDT::REAL_VAL && oSecond.GetType() <= CDT::REAL_VAL)
								{
									W_FLOAT dTMP = oFirst.GetFloat() - oSecond.GetFloat();
									if      (dTMP < 0.0) { iFlags = FL_LT | FL_NE; }
									else if (dTMP > 0.0) { iFlags = FL_GT | FL_NE; }
									else                 { iFlags = FL_EQ; }

#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%f`, `%f`)\n", oFirst.GetFloat(), oSecond.GetFloat());
HL_RST;
#endif


									if (oFirst.GetType() == CDT::INT_VAL)
									{
										UINT_64 iTMP = oFirst.GetInt();
										if ((iTMP % 2) == 1) { iFlags |= FL_ODD; }
									}
								}
								else
								{
									CCHAR_P sAR = oFirst.GetString().c_str();
									CCHAR_P sBR = oSecond.GetString().c_str();
									INT_32 iResult = strcmp(sAR, sBR);
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%s`, `%s`)\n", sAR, sBR);
HL_RST;
#endif
									if      (iResult < 0.0) { iFlags = FL_LT; }
									else if (iResult > 0.0) { iFlags = FL_GT; }
									else                    { iFlags = FL_EQ; }
								}
							}
						}
						++iIP;
					}
					// Conditional jump
					else if (iHiOpcode == JXX)
					{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X JXX       0x%08X (0x%08X <-> 0x%08X)\n", iIP, aCode[iIP].argument, iLoLoOpcode, iFlags);
#endif
						if (iExecutedSteps >= iMaxSteps) { throw ExecutionLimitReached(iIP); }

						const UINT_32 iNewIP = aCode[iIP].argument;
						if (iNewIP >= iCodeLength) { throw CodeSegmentOverrun(iIP); }

						if ((iLoLoOpcode & iFlags) == iLoLoOpcode) { iIP = aCode[iIP].argument; }
						else { ++iIP; }
					}
					// Check existence of operand
					else if (iHiOpcode == EXIST)
					{

#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X EXIST     ", iIP);
#endif
						if (iLoLoOpcode == ARG_SRC_STACK)
						{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](%c)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).Defined() ? 't':'f');
HL_RST;
#endif
							// Defined
							if (oVMArgStack.GetTopElement(aCode[iIP].argument).Defined()) { iFlags = FL_EQ; }
							// Undefined
							else { iFlags = FL_NE; }
						}
						else if (iLoLoOpcode <= ARG_SRC_LASTREG)
						{
#ifdef _DEBUG
fprintf(stderr, "%cR (`%s`)\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
							// Defined
							if (oRegs[iLoLoOpcode].Defined()) { iFlags = FL_EQ; }
							// Undefined
							else                              { iFlags = FL_NE; }
						}
						else { throw IllegalOpcode(iIP, iOpCode); }

						++iIP;
					}
					// Move ONLY to registers from registers or from stack
					else if (iHiOpcode == MOV)
					{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOV       ", iIP);
#endif
						if (iLoHiOpcode <= ARG_DST_LASTREG)
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
#ifdef _DEBUG
fprintf(stderr, "%cR, ", CHAR_8(iDstReg + 'A'));
#endif
							// Register-to-register
							if      (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR == `%s`\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
								oRegs[iDstReg] = oRegs[iLoLoOpcode];
							}
							// Stack element-to-register
							else if (iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d] = `%s`\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								oRegs[iDstReg] = oVMArgStack.GetTopElement(aCode[iIP].argument);
							}
							// Integer-to-register
							else if (iLoLoOpcode == ARG_SRC_INT)
							{
#ifdef _DEBUG
fprintf(stderr, "INT POS: %d (VAL: %d)\n", aCode[iIP].argument, INT_32(oMemoryCore.static_data.GetInt(aCode[iIP].argument)));
HL_RST;
#endif
								oRegs[iDstReg] = oMemoryCore.static_data.GetInt(aCode[iIP].argument);
							}
							// Float-to-register
							else if (iLoLoOpcode == ARG_SRC_FLOAT)
							{
#ifdef _DEBUG
fprintf(stderr, "FLOAT POS: %d (VAL: %f)\n", aCode[iIP].argument, oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
HL_RST;
#endif
								oRegs[iDstReg] = oMemoryCore.static_data.GetFloat(aCode[iIP].argument);
							}
							// String-to-register
							else if (iLoLoOpcode == ARG_SRC_STR)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "STRING POS: %d (VAL: `%s`)\n", aCode[iIP].argument, szTMP);
HL_RST;
#endif
								oRegs[iDstReg] = std::string(szTMP, iDataSize);
							}
							else { throw IllegalOpcode(iIP, iOpCode); }
						}
						else { throw IllegalOpcode(iIP, iOpCode); }

						++iIP;
					}
					// Indirect operations works ONLY with registers AR - HR
					//
					// Move indirect ARRAY to register
					else if (iHiOpcode == MOVIINT)
					{
						UINT_32 iDstReg = (iLoHiOpcode >> 8);
						oRegs[iDstReg] = oRegs[iLoLoOpcode].GetCDT(aCode[iIP].argument);
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVIINT   %cR, %cR[%d] (`%s`)", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), aCode[iIP].argument, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
					}
					// Move indirect HASH to register
					else if (iHiOpcode == MOVISTR)
					{
						if (iLoLoOpcode > ARG_SRC_LASTREG) { throw IllegalOpcode(iIP, iOpCode); }

						UINT_32 iDstReg = (iLoHiOpcode >> 8);

						if (iDstReg > ARG_SRC_LASTREG) { throw IllegalOpcode(iIP, iOpCode); }

						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);

						INT_32 iCDTExist = 0;
						oRegs[iDstReg] = oRegs[iLoLoOpcode].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist);

						// Found
						if (iCDTExist == 0) { iFlags = FL_EQ; }
						// Not Found
						else                { iFlags = FL_NE; }
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVISTR   %cR, %cR[%s] (`%s`)\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), szTMP, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					// Move register to indirect ARRAY
					else if (iHiOpcode == IMOVINT)
					{
						UINT_32 iDstReg = (iLoHiOpcode >> 8);
						oRegs[iLoLoOpcode][aCode[iIP].argument] = oRegs[iDstReg];
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X IMOVINT   %cR[%d], %cR[%d] (`%s`)\n", iIP, CHAR_8(iDstReg + 'A'), aCode[iIP].argument, CHAR_8(iLoLoOpcode + 'A'), aCode[iIP].argument, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					// Move register to indirect HASH
					else if (iHiOpcode == IMOVSTR)
					{
						UINT_32 iDstReg = (iLoHiOpcode >> 8);
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
						oRegs[iLoLoOpcode][std::string(szTMP, iDataSize)] = oRegs[iDstReg];
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X IMOVSTR   %cR[%s], %cR (`%s`)", iIP, CHAR_8(iDstReg + 'A'), szTMP, CHAR_8(iLoLoOpcode + 'A'), oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
						++iIP;
					}
					else if (iHiOpcode == MOVSIZE)
					{
						if (iLoHiOpcode <= ARG_DST_LASTREG)
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							// Register-to-register
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVSIZE   %cR, %cR.Size() (%d)\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), UINT_32(oRegs[iLoLoOpcode].Size()));
HL_RST;
#endif
								oRegs[iDstReg] = oRegs[iLoLoOpcode].Size();
							}
							else { throw IllegalOpcode(iIP, iOpCode); }
						}
						else { throw IllegalOpcode(iIP, iOpCode); }
						++iIP;
					}
					else if (iHiOpcode == MOVIREG)
					{
						if (iLoHiOpcode <= ARG_DST_LASTREG)
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							// Register-to-register
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
								UINT_32 iArgNum = aCode[iIP].argument;
								if (iArgNum > ARG_SRC_LASTREG) { throw IllegalOpcode(iIP, iOpCode); }
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X MOVIREG   %cR, %cR[%cR] (%d)\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), CHAR_8(iArgNum + 'A'), INT_32(oRegs[iArgNum].GetInt()));
HL_RST;
#endif
								oRegs[iDstReg] = oRegs[iLoLoOpcode][oRegs[iArgNum].GetInt()];
							}
							else { throw IllegalOpcode(iIP, iOpCode); }
						}
						else { throw IllegalOpcode(iIP, iOpCode); }
						++iIP;
					}
					// Push value to stack
					else if (iHiOpcode == PUSH)
					{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X PUSH      ", iIP);
#endif
						// From static text segment to stack
						if (iLoLoOpcode == ARG_SRC_STR)
						{
							UINT_32 iDataSize = 0;
							CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "STRING POS: %d (VAL: `%s`)\n", aCode[iIP].argument, szTMP);
HL_RST;
#endif
							oVMArgStack.PushElement(std::string(szTMP, iDataSize));
						}
						// From static data segment to stack (integer value)
						else if (iLoLoOpcode == ARG_SRC_INT)
						{
#ifdef _DEBUG
fprintf(stderr, "INT POS: %d (VAL: %d)\n", aCode[iIP].argument, INT_32(oMemoryCore.static_data.GetInt(aCode[iIP].argument)));
HL_RST;
#endif
							oVMArgStack.PushElement(oMemoryCore.static_data.GetInt(aCode[iIP].argument));
						}
						// From static data segment to stack (float value)
						else if (iLoLoOpcode == ARG_SRC_FLOAT)
						{
#ifdef _DEBUG
fprintf(stderr, "FLOAT POS: %d (VAL: %f)\n", aCode[iIP].argument, oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
HL_RST;
#endif
							oVMArgStack.PushElement(oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
						}
						// From register to stack
						else if (iLoLoOpcode <= ARG_SRC_LASTREG)
						{
							oVMArgStack.PushElement(oRegs[iLoLoOpcode]);
#ifdef _DEBUG
fprintf(stderr, "%cR\n", CHAR_8(iLoLoOpcode + 'A'));
HL_RST;
#endif
						}
						// Indirect operations works ONLY with registers AR - HR
						else if (iLoLoOpcode == ARG_SRC_IND_VAL)
						{
#ifdef _DEBUG
fprintf(stderr, "%cR[%d] (`%s`)\n", CHAR_8((iLoHiOpcode >> 8) + 'A'), aCode[iIP].argument, oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								oVMArgStack.PushElement(oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument));
							}
							else { throw IllegalOpcode(iIP, iOpCode); }

						}
						else if (iLoLoOpcode == ARG_SRC_IND_STR)
						{
							UINT_32 iDataSize = 0;
							CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "%cR[\"%s\"] (`%s`)\n", CHAR_8((iLoHiOpcode >> 8) + 'A'), szTMP, oRegs[iLoHiOpcode >> 8].GetCDT(std::string(szTMP, iDataSize)).GetString().c_str());
HL_RST;
#endif
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								INT_32 iCDTExist = 0;
								oVMArgStack.PushElement(oRegs[iLoHiOpcode >> 8].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist));

								// Found
								if (iCDTExist == 0) { iFlags = FL_EQ; }
								// Not Found
								else                { iFlags = FL_NE; }
							}
							else { throw IllegalOpcode(iIP, iOpCode); }
						}
						// Illegal operation
						else { throw IllegalOpcode(iIP, iOpCode); }
						++iIP;
					}
					// Clear register
					else if (iHiOpcode == CLEAR)
					{
						if (iLoHiOpcode <= ARG_DST_LASTREG)
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							oRegs[iDstReg] = CDT();
						}
						else { throw IllegalOpcode(iIP, iOpCode); }
					}
					// Illegal opcode called
					else { throw IllegalOpcode(iIP, iOpCode); }
				} // End default switch branch
			} // End switch
			++iExecutedSteps;
		}
	}
	catch (StackOverflow  &e) { throw StackOverflow(iIP);  }
	catch (StackUnderflow &e) { throw StackUnderflow(iIP); }

return 0;
}

//
// Run one instruction
//
//INT_32 VM::Step(VMInstruction * aCode, const UINT_32 & iCodeLength, UINT_32 & iIP)
//{
//return -1;
//}

//
// Reset virtual machine state
//
INT_32 VM::Reset()
{
	AR = CDT();
	BR = CDT();
	CR = CDT();
	DR = CDT();
	ER = CDT();
	FR = CDT();
	GR = CDT();
	HR = CDT();

	iFlags = 0;

	oVMArgStack.Reset();
	oVMCodeStack.Reset();

return 0;
}

//
// A destructor
//
VM::~VM() throw()
{
	if (aCallTranslationMap != 0) { delete [] aCallTranslationMap; }

}

} // namespace CTPP
// End.
